/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.OnEvent;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.CorrelationSetDefinition;
import org.ow2.orchestra.facade.def.PartnerLinkDefinition;
import org.ow2.orchestra.facade.def.VariableDefinition;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.CopyFullDefinition;
import org.ow2.orchestra.facade.def.full.OnEventFullDefinition;
import org.ow2.orchestra.facade.def.full.ScopeActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.CopyFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.OnEventFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.ScopeActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.CorrelationSetDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.PartnerLinkDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.VariableDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.parsing.binding.exception.DuplicateVariableException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.util.XmlUtil;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.OrchestraFaultHandler;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */

public class ScopeBinding extends ActivityWithFCTHBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ScopeBinding.class.getName());

  public ScopeBinding() {
    super("scope", ActivityType.SCOPE);
  }

  public ScopeBinding(final String tagName, final ActivityType activityType) {
    super(tagName, activityType);
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *          dom element to parse
   * @param parse
   *          parse
   * @param parser
   *          parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    try {
      return this.parse(element, null, parse, parser);
    } catch (final DuplicateVariableException e) {
      throw new Error();
    }
  }

  /**
   * Factory method (reused by {@link ProcessBinding})
   *
   * @return {@link org.ow2.orchestra.definition.activity.AbstractBpelActivity}
   *         associated with this node (must extend {@link Scope})
   */
  protected Scope createScopeActivity() {
    return new Scope();
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *          dom element to parse
   * @param additionnal
   *          variables to add
   * @param parse
   *          parse
   * @param parser
   *          parser
   * @return the build behaviour!
   * @throws DuplicateVariableException
   *           if duplicated variabled
   */
  public Object parse(final Element element, final Set<Variable> moreVariables, final Parse parse, final Parser parser)
    throws DuplicateVariableException {
    if (ScopeBinding.log.isLoggable(Level.FINE)) {
      ScopeBinding.log.fine("parsing element = " + element);
    }

    final Scope scope = this.createScopeActivity();
    this.setEnclosingScope(scope, parse);
    this.parseStandardAttributes(element, scope, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ScopeActivityFullDefinition activityFullDefinition = (ScopeActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();

    this.parseStandardElements(element, scope, parse);

    final WsdlsInfos wsdlInfos = parse.findObject(BpelProcess.class).getWsdlInfos();

    // parse attributes
    final String isolatedAttr = XmlUtil.attribute(element, "isolated");
    if (isolatedAttr != null) {
      final boolean isolated = BpelUtil.booleanValue(isolatedAttr);
      scope.setIsolated(isolated);
      if (isolated) {
        parse.addProblem("isolated attribute not yet supported on scope activity");
      }
    }

    boolean defaultExitOnStandardFault = false;
    if (scope.getEnclosingScope() != null) {
      defaultExitOnStandardFault = scope.getEnclosingScope().isExitOnStandardFault();
    }
    final boolean exitOnStandardFault = BpelUtil.booleanValue(XmlUtil.attribute(element, "exitOnStandardFault"),
      defaultExitOnStandardFault);
    scope.setExitOnStandardFault(exitOnStandardFault);
    if (exitOnStandardFault) {
      ScopeBinding.log.warning("exitOnStandardFault attribute not yet supported on scope activity");
    }
    activityFullDefinition.setExitOnStandardFault(exitOnStandardFault);

    // partnerLinks
    final Element partnerLinksElement = XmlUtil.element(element, "partnerLinks");
    if (partnerLinksElement != null) {
      final List<Element> partnerLinkElements = XmlUtil.elements(partnerLinksElement, "partnerLink");
      this.parsePartnerLinks(partnerLinkElements, wsdlInfos, scope, parse);
    }

    // messageExchanges
    final Element messageExchangesElement = XmlUtil.element(element, "messageExchanges");
    if (messageExchangesElement != null) {
      final List<Element> messageExchangeElements = XmlUtil.elements(messageExchangesElement, "messageExchange");
      this.parseMessageExchanges(messageExchangeElements, scope, parse);
    }

    // variables
    final Element variablesElement = XmlUtil.element(element, "variables");
    if (variablesElement != null) {
      final List<Element> variableElements = XmlUtil.elements(variablesElement, "variable");
      this.parseVariables(variableElements, scope, parse);
    }

    // add more variables
    if (moreVariables != null) {
      Map<String, Variable> variables = scope.getVariables();
      final List<VariableDefinition> variableDefinitionList = new ArrayList<VariableDefinition>();

      if (variables == null) {
        variables = new HashMap<String, Variable>();
      }
      for (final Variable var : moreVariables) {
        if (variables.put(var.getName(), var) != null) {
          throw new DuplicateVariableException();
        } else {
          // Add variable definition to the scope definition list
          final VariableDefinitionImpl variableDefinition = new VariableDefinitionImpl(var.getElement(), var.getMessageType(),
            var.getName(), var.getXmlType());
          variableDefinitionList.add(variableDefinition);
        }
      }
      scope.setVariables(variables);
      activityFullDefinition.setVariableDefinition(variableDefinitionList);
    }

    // correlationSets
    final Element correlationSetsElement = XmlUtil.element(element, "correlationSets");
    if (correlationSetsElement != null) {
      final List<Element> correlationSetElements = XmlUtil.elements(correlationSetsElement, "correlationSet");
      this.parseCorrelationSets(correlationSetElements, scope, wsdlInfos, parse);
    }

    final NodeImpl scopeNode = this.createActivityNodeWithChildren(element, parse, parser, scope);

    scope.setScopeNode(scopeNode);

    try {
      parse.pushObject(scope);
      parse.pushObject(scopeNode);

      // Analyze the terminationHandler
      this.parseTermination(element, parse, parser, scopeNode, activityFullDefinition);

      // Analyze the compensationHandler
      this.parseCompensation(element, parse, parser, scopeNode, activityFullDefinition);

      // Analyze the faultHandlers
      final Element faultHandlersElement = XmlUtil.element(element, "faultHandlers");
      if (faultHandlersElement != null) {
        parse.pushObject(faultHandlersElement);

        try {
          final int elements = this.parseCatchAndCatchAll(faultHandlersElement, parse, parser, scopeNode, activityFullDefinition);

          // Static Analysis SA00080
          if (elements < 1) {
            final String message = StaticAnalysisFault.getMessage("SA00080", "  The <faultHandlers> element for scope \""
              + scope.getName() + "\" does not have any <catch> or <catchAll> elements.");
            parse.addProblem(message, new StaticAnalysisException("SA00080"));
            return null;
          }
        } finally {
          parse.popObject();
        }
      } else {
        // Add the default faultHandler node in order to respect indices
        this.addDefaultFH(parse, scopeNode);

        // Since we did not call parseCatchAndCatchAll, we need to add the
        // exception handler to the scopeNode manually
        scopeNode.addExceptionHandler(new OrchestraFaultHandler());
      }

      // eventHandlers
      final Element eventHandlersElement = XmlUtil.element(element, "eventHandlers");
      if (eventHandlersElement != null) {

        parse.addProblem("eventHandlers not yet supported on scope activity");

        // TODO : uncomment the foolowing part to activate eventHandlers
        // parsing. The code must be ckecked before, as the entire parsing is
        // not necessary done. Plus, an EventHandlerActivity has to be set up.
        // Also activate the corresponding tests for SA.

        //
        // parse.pushObject(eventHandlersElement);
        //
        // try {
        // final int elements = this.parseEventHandlers(eventHandlersElement,
        // scope, parse, parser, scopeNode);
        //
        // if (elements < 1) {
        // final String message = StaticAnalysisFault.getMessage("SA00083",
        // "The <eventHandlers> element does not have any <onEvent> or <onAlarm> element");
        // parse.addProblem(message, new StaticAnalysisException("SA00083"));
        // }
        //
        // } finally {
        // parse.popObject();
        // }
      }

    } finally {
      parse.popObject();
      parse.popObject();
    }

    final List<Scope> immediatelyEnclosingScopes = this.getImmediatelyEnclosingScopes(scopeNode);
    for (final Scope scope1 : immediatelyEnclosingScopes) {
      if (scope1.getName() != null && !XmlConstants.BPEL_DEFAULT_ACTIVITY_NAME.equals(scope1.getName())) {
        for (final Scope scope2 : immediatelyEnclosingScopes) {
          if (!scope1.getUuid().equals(scope2.getUuid())) {
            if (scope1.getName().equals(scope2.getName())) {
              final String message = StaticAnalysisFault.getMessage("SA00092", "  The scope name \"" + scope1.getName()
                + "\" is used twice in the immediately enclosing scopes of scope \"" + scope.getName() + "\"");
              parse.addProblem(message, new StaticAnalysisException("SA00092"));
            }
          }
        }
      }
    }

    return scopeNode;
  }

  /**
   * A scope A is considered to immediately enclose another scope B, if B is
   * enclosed in A and B is not enclosed in any other scope or FCT-handler that
   * is itself enclosed in the outer scope A. Other structured activities (e.g.
   * &lt;sequence&gt; or &lt;forEach&gt;) and event handlers enclosed in A do
   * not affect the concept of immediate enclosure. This definition includes
   * scopes that result from the &lt;invoke&gt; shorthand notation for fault
   * handlers and compensation handlers.<br/>
   * <br/>
   * <p align="right"><b><i>wsbpel-v2.0-OS, page 123, paragraph 2</i></b></p>
   */
  private List<Scope> getImmediatelyEnclosingScopes(final NodeImpl node) {
    final List<Scope> result = new ArrayList<Scope>();
    if (node.getNodes() != null) {
      for (final Node subNode : node.getNodes()) {
        final NodeImpl subNodeImpl = (NodeImpl) subNode;
        final AbstractActivity subNodeActivity = (AbstractActivity) subNodeImpl.getBehaviour();
        final ActivityType subNodeActivityType = subNodeActivity.getType();
        if (subNodeActivityType == ActivityType.SCOPE) {
          result.add((Scope) subNodeActivity);
        } else if (subNodeActivityType != ActivityType.CATCH_HANDLER && subNodeActivityType != ActivityType.CATCHALL_HANDLER
          && subNodeActivityType != ActivityType.TERMINATION_HANDLER && subNodeActivityType != ActivityType.COMPENSATION_HANDLER) {
          result.addAll(this.getImmediatelyEnclosingScopes(subNodeImpl));
        }
      }
    }
    return result;
  }

  /**
   * Parses all partnerLink elements of the fiven scope.
   *
   * @param partnerLinkElements
   *          list of all partnerLink elements of this scope
   * @param wsdlInfos
   *          wsdlInfos
   * @param scope
   *          current scope
   */
  private void parsePartnerLinks(final List<Element> partnerLinkElements, final WsdlsInfos wsdlInfos, final Scope scope,
    final Parse parse) {
    final Map<String, PartnerLink> partnerLinks = new HashMap<String, PartnerLink>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<PartnerLinkDefinition> partnerLinkDefinitionList = new ArrayList<PartnerLinkDefinition>();

    if (partnerLinkElements != null) {
      for (final Element partnerLinkElement : partnerLinkElements) {
        // variables use to construcut the PartnerLinkDefinition
        String myRoleNameDefinition = null;
        String partnerRoleNameDefinition = null;
        QName partnerLinkTypeDefinition;

        final PartnerLink partnerLink = new PartnerLink();
        final String name = XmlUtil.attribute(partnerLinkElement, "name");
        if (partnerLinks.containsKey(name)) {
          // Static Analysis SA00018
          final String message = StaticAnalysisFault.getMessage("SA00018", "  PartnerLink " + name + " defined twice in scope "
            + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00018"));
        } else {
          partnerLink.setName(name);

          final QName partnerLinkTypeName = XmlUtil.attributeQName(partnerLinkElement, "partnerLinkType");
          final PartnerLinkType partnerLinkType = wsdlInfos.getPartnerLinkType(partnerLinkTypeName);

          partnerLinkTypeDefinition = partnerLinkTypeName;
          boolean roleDef = false;
          if (partnerLinkType != null) {
            final String myRoleName = XmlUtil.attribute(partnerLinkElement, "myRole");
            if (myRoleName != null) {
              final QName myRolePortTypeQName = partnerLinkType.getPortTypeQName(myRoleName);
              partnerLink.setMyRolePortTypeQName(myRolePortTypeQName);
              roleDef = true;

              myRoleNameDefinition = myRoleName;
            }
            final String partnerRoleName = XmlUtil.attribute(partnerLinkElement, "partnerRole");
            if (partnerRoleName != null) {
              final QName partnerRolePortTypeQName = partnerLinkType.getPortTypeQName(partnerRoleName);
              partnerLink.setPartnerRolePortTypeQName(partnerRolePortTypeQName);
              roleDef = true;

              partnerRoleNameDefinition = partnerRoleName;
            }
            final String initializePartnerRole = XmlUtil.attribute(partnerLinkElement, "initializePartnerRole");
            if (initializePartnerRole != null) {
              if (partnerRoleName != null) {
                partnerLink.setInitializePartnerRole(initializePartnerRole.equals("true"));
              } else {
                // SA00017
                final String message = StaticAnalysisFault.getMessage("SA00017", "  PartnerLink " + name
                  + " has initializePartnerRole attribute used and does not have a partner role.");
                parse.addProblem(message, new StaticAnalysisException("SA00017"));
              }
            }
          }
          if (!roleDef) {
            // SA00016
            final String message = StaticAnalysisFault.getMessage("SA00016", "  PartnerLink " + name
              + " should specify myRole or partnerRole or both");
            parse.addProblem(message, new StaticAnalysisException("SA00016"));
          } else {
            partnerLinks.put(name, partnerLink);

            // Add variable definition to the scope definition list
            final PartnerLinkDefinitionImpl partnerLinkDefinition = new PartnerLinkDefinitionImpl(myRoleNameDefinition,
              partnerLink.getName(), partnerLinkTypeDefinition, partnerRoleNameDefinition, partnerLink.getInitializePartnerRole());

            partnerLinkDefinitionList.add(partnerLinkDefinition);
          }
        }
      }
    }
    scope.setPartnerLinks(partnerLinks);
    ((ScopeActivityFullDefinition) activityFullDefinition).setPartnerLinkDefinition(partnerLinkDefinitionList);
  }

  /**
   * Parses all messageExchanges of the given scope.
   *
   * @param messageExchangeElements
   *          list of all messageExchange elements of the given scope
   * @param scope
   *          scope
   */
  private void parseMessageExchanges(final List<Element> messageExchangeElements, final Scope scope, final Parse parse) {
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();

    final Set<String> messageExchanges = new HashSet<String>();
    if (messageExchanges != null) {
      for (final Element messageExchangeElement : messageExchangeElements) {
        final String name = XmlUtil.attribute(messageExchangeElement, "name");
        messageExchanges.add(name);
      }
    }
    scope.setMessageExchanges(messageExchanges);
    ((ScopeActivityFullDefinition) activityFullDefinition).setMessageExchange(messageExchanges);
  }

  /**
   * Parses all variables of the given scope.
   *
   * @param variableElements
   *          list of all variable elements of the given scope
   * @param scope
   *          scope
   * @param parse
   *          parse
   */
  private void parseVariables(final List<Element> variableElements, final Scope scope, final Parse parse) {
    final Map<String, Variable> variables = new HashMap<String, Variable>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<VariableDefinition> variableDefinitionList = new ArrayList<VariableDefinition>();

    if (variableElements != null) {
      for (final Element variableElement : variableElements) {
        final Variable variable = new Variable();
        final String name = XmlUtil.attribute(variableElement, "name");
        if (variables.containsKey(name)) {
          // Static Analysis SA00023
          final String message = StaticAnalysisFault.getMessage("SA00023", "  Variable " + name + " defined twice in scope "
            + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00023"));
        } else {
          variable.setName(name);
          variable.setMessageType(XmlUtil.attributeQName(variableElement, "messageType"));
          variable.setXmlType(XmlUtil.attributeQName(variableElement, "type"));
          variable.setElement(XmlUtil.attributeQName(variableElement, "element"));

          if (!(((variable.getElement() != null) && (variable.getMessageType() == null) && (variable.getXmlType() == null))
            || ((variable.getElement() == null) && (variable.getMessageType() != null) && (variable.getXmlType() == null)) || ((variable
            .getElement() == null)
            && (variable.getMessageType() == null) && (variable.getXmlType() != null)))) {
            // Static Analysis SA00025
            final String message = StaticAnalysisFault.getMessage("SA00025", "  Type of variable " + name
              + " not defined properly.");
            parse.addProblem(message, new StaticAnalysisException("SA00025"));
          } else {

            final Element fromElement = XmlUtil.element(variableElement, "from");
            if (fromElement != null) {
              final CopyFullDefinition copyFullDefinition = new CopyFullDefinitionImpl();
              final Copy virtualCopy = new Copy();
              virtualCopy.setKeepSrcElementName(true);
              AssignBinding.parseFrom(fromElement, virtualCopy, parse, scope, copyFullDefinition);
              variable.setVirtualCopy(virtualCopy);
              final To to = new To();
              to.setVariable(name);
              to.setVariant(Variant.VAR);
              virtualCopy.setTo(to);
              scope.addVariableToBeInitialized(name);
            }
            variables.put(name, variable);

            // Add variable definition to the scope definition list
            final VariableDefinitionImpl variableDefinition = new VariableDefinitionImpl(variable.getElement(), variable
              .getMessageType(), variable.getName(), variable.getXmlType());
            variableDefinitionList.add(variableDefinition);
          }
        }
      }
    }
    scope.setVariables(variables);
    ((ScopeActivityFullDefinition) activityFullDefinition).setVariableDefinition(variableDefinitionList);
  }

  /**
   * Parses all correlationSet of the given scope.
   *
   * @param correlationSetElements
   *          list of all correlationSet elements of the given scope
   * @param scope
   *          scope
   * @param wsdlInfos
   *          wsdlInfos
   */
  private void parseCorrelationSets(final List<Element> correlationSetElements, final Scope scope, final WsdlsInfos wsdlInfos,
    final Parse parse) {
    final Map<String, CorrelationSet> correlationSets = new HashMap<String, CorrelationSet>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<CorrelationSetDefinition> correlationSetDefinitionList = new ArrayList<CorrelationSetDefinition>();

    if (correlationSetElements != null) {
      for (final Element correlationSetElement : correlationSetElements) {
        final CorrelationSet cs = new CorrelationSet();
        final String name = XmlUtil.attribute(correlationSetElement, "name");
        if (correlationSets.containsKey(name)) {
          // Static Analysis SA00044
          final String message = StaticAnalysisFault.getMessage("SA00044", "  CorrelationSet " + name
            + " defined twice in scope " + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00044"));
        } else {
          cs.setName(name);
          final String propertiesSt = XmlUtil.attribute(correlationSetElement, "properties");
          final Map<QName, Property> properties = new HashMap<QName, Property>();
          final String[] propertyQNames = propertiesSt.split(" ");

          final List<QName> qNameList = new ArrayList<QName>();
          for (String prop : propertyQNames) {
            prop = prop.trim();
            if (!prop.equals("")) {
              final QName propertyQN = XmlUtil.getQNameFromString(correlationSetElement, prop);
              final Property property = wsdlInfos.getProperty(propertyQN);
              properties.put(propertyQN, property);
              qNameList.add(propertyQN);
            }
          }
          cs.setProperties(properties);
          correlationSets.put(name, cs);

          final CorrelationSetDefinitionImpl correlationSetDefinition = new CorrelationSetDefinitionImpl(name, qNameList);
          correlationSetDefinitionList.add(correlationSetDefinition);
        }
      }
    }
    scope.setCorrelationSets(correlationSets);
    ((ScopeActivityFullDefinition) activityFullDefinition).setCorrelationSetDefinition(correlationSetDefinitionList);
  }

  private int parseEventHandlers(final Element eventHandlersElement, final Scope scope, final Parse parse, final Parser parser,
    final NodeImpl scopeNode) {

    int onEventElementsNumber = 0;
    int onAlarmElementsNumber = 0;

    List<OnEvent> onEvents;

    final List<Element> onEventElements = XmlUtil.elements(eventHandlersElement, "onEvent");

    if (onEventElements != null) {
      onEventElementsNumber = onEventElements.size();

      onEvents = new ArrayList<OnEvent>();

      for (final Element onEventElement : onEventElements) {
        final OnEventFullDefinition onEventFullDefinition = new OnEventFullDefinitionImpl();

        final OnEvent onEvent = this.parseOnEvent(onEventElement, scope, parse, parser, onEventFullDefinition);
        onEvents.add(onEvent);

        // onMessageFullDefinitionList.add(onMessageFullDefinition);

      }
    }

    final List<Element> onAlarmElements = XmlUtil.elements(eventHandlersElement, "onAlarm");
    if (onAlarmElements != null) {
      onAlarmElementsNumber = onAlarmElements.size();
    }

    return onAlarmElementsNumber + onEventElementsNumber;
  }

  private OnEvent parseOnEvent(final Element element, final Scope scope, final Parse parse, final Parser parser,
    final OnEventFullDefinition onEventFullDefinition) {
    if (ScopeBinding.log.isLoggable(Level.FINE)) {
      ScopeBinding.log.fine("parsing element = " + element);
    }
    final OnEvent onEvent = new OnEvent();

    final String variable = XmlUtil.attribute(element, "variable");
    final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    final String operation = XmlUtil.attribute(element, "operation");
    final String portType = XmlUtil.attribute(element, "portType");
    final String messageExchange = XmlUtil.attribute(element, "messageExchange");

    if (messageExchange != null) {
      onEvent.setMessageExchange(messageExchange);
      onEventFullDefinition.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    final PartnerLink partnerLink = this.findPartnerLinkDefinition(parse, partnerLinkName);
    if (partnerLink == null) {
      parse.addProblem("No partnerLink " + partnerLinkName + " defined.");
      return null;
    }
    final QName plportType = partnerLink.getMyRolePortTypeQName();
    if (plportType == null) {
      parse.addProblem("No myRole defined for partnerLink: " + partnerLinkName);
      return null;
    }
    if (portType != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        final String message = StaticAnalysisFault.getMessage("SA00005", "  onMessage " + onEvent
          + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // Definition
    onEventFullDefinition.setPortType(onEvent.getPortTypeQName());
    //
    // // fromParts
    // Element fromPartsElement = XmlUtil.element(element, "fromParts");
    // if (fromPartsElement != null) {
    // if (onEvent.getVariable() != null) {
    // // SA00055
    // String message = StaticAnalysisFault.getMessage("SA00063", "  onMessage "
    // + onEvent
    // + " has specified both a Variable and a fromPart");
    // parse.addProblem(message, new StaticAnalysisException("SA00063"));
    // }
    // List<Element> fromPartElements = XmlUtil.elements(fromPartsElement,
    // "fromPart");
    // onEvent.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onEvent,
    // parse));
    // List<Copy> inCopies = ReceiveBinding.getInCopies(fromPartElements,
    // pick.getEnclosingScope(), parse, onEventFullDefinition);
    // if (inCopies.size() != 0) {
    // onEvent.setInCopies(inCopies);
    // }
    // }

    // Definition
    final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onEventFullDefinition);

    // correlations

    // FIXME : SA00088. Here, only the parent scopes are checked for the parsing
    final Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      final List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      onEvent.setCorrelations(ReceiveBinding.parseCorrelations(correlationElements, scope, parse, onEventFullDefinition));
    }
    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    // onEvent.setActivity(this.parseChildActivity(element, parse, parser));

    defKeeper.popParent();

    // Need to declare a variable that fits the needs ...
    final String messageType = XmlUtil.attribute(element, "messageType");
    final String elementAttribute = XmlUtil.attribute(element, "element");

    onEvent.setVariable(variable);
    onEvent.setPartnerLinkName(partnerLinkName);
    onEvent.setOperationKey(new OperationKey(bpelProcess.getQName(), plportType, operation));

    // Definition
    onEventFullDefinition.setOperation(operation);
    onEventFullDefinition.setPartnerLink(partnerLinkName);
    onEventFullDefinition.setVariable(variable);

    // check attributes
    if (variable != null) {
      // either both are set or none
      if ((messageType != null && elementAttribute != null) || (messageType == null && elementAttribute == null)) {
        // SA00090
        final String message = StaticAnalysisFault.getMessage("SA00090", "  onEvent " + onEvent
          + " has specified both an element attribute and a messageType one.");
        parse.addProblem(message, new StaticAnalysisException("SA00090"));
      }

      // Add input variable to onEvent enclosed scope
      // This is done by giving additional variables as input parameter
      // during scope parsing.

      // Fetch enclosed scope ...
      final List<Element> activityElements = XmlUtil.elementsQName(element, Collections.singleton(new QName(
        XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, "scope")));
      if (activityElements.size() == 0) {
        parse.addProblem("No scope was specified under this onEvent element.");
        return null;
      }
      // ... in childrenElement
      final Element childrenElement = activityElements.get(0);

      // declare the input variable
      final Variable variableInput = new Variable();
      variableInput.setName(onEvent.getVariable());

      variableInput.setMessageType(XmlUtil.attributeQName(element, "messageType"));
      variableInput.setXmlType(XmlUtil.attributeQName(element, "type"));
      variableInput.setElement(XmlUtil.attributeQName(element, "element"));

      // Get binding of enclosed scope
      final ScopeBinding binding = (ScopeBinding) parser.getBinding(childrenElement, BpelParser.CATEGORY_ACTIVITY);
      // and parse the scope with method which allow to define some additional
      // variables

      try {
        defKeeper.pushParent();
        onEvent.setActivity((NodeImpl) binding.parse(childrenElement, Collections.singleton(variableInput), parse, parser));
      } catch (final DuplicateVariableException e) {
        // SA00086
        final String message = StaticAnalysisFault.getMessage("SA00086",
          "enclosed scope declare a variable with input variable name of onEvent activity");
        parse.addProblem(message, new StaticAnalysisException("SA00086"));
      } finally {
        defKeeper.popParent();
      }

    } else {
      if (messageType != null || elementAttribute != null) {
        // SA00090
        final String message = StaticAnalysisFault.getMessage("SA00090", "  onEvent " + onEvent
          + " has specified either an element attribute or a messageType one, but no variable attribute is set.");
        parse.addProblem(message, new StaticAnalysisException("SA00090"));
      }
    }

    return onEvent;
  }
  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new ScopeActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
