/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.UUID;
import java.util.concurrent.TimeoutException;

import javax.transaction.Synchronization;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.Transaction;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.var.Message;

public class MessageCarrier {

  private final Object lock = new Object();

  private Message message;

  private boolean hasMessage = false;

  private final String uuid;

  public MessageCarrier() {
    this.uuid = UUID.randomUUID().toString();
  }

  /**
   * Wait until the message has arrived.
   * @return message
   */
  public Message getMessage() {
    try {
      return this.getMessage(0);
    } catch (final TimeoutException e) {
      Misc.unreachableStatement("No timeout shuld occur");
      return null;
    }
  }

  /**
   * Wait until the message has arrived or timeout occurs.
   * @param timeout timeout in millisecond.
   * If timeout is set to 0 then this method has the same behaviour as {@link #getMessage()}
   * @return message
   * @throws TimeoutException thrown if no message has arrived when timeout occurs.
   */
  public Message getMessage(final long timeout) throws TimeoutException {
    synchronized (this.lock) {
      while (!this.hasMessage) {
        try {
          this.lock.wait(timeout);
          if (!this.hasMessage) {
            throw new TimeoutException("Timeout occured while waiting for message");
          }
        } catch (final InterruptedException e) {
          throw new OrchestraRuntimeException(e);
        }
      }
      return this.message;
    }
  }

  private void replyMessage(final Message messageToSend) {
    synchronized (this.lock) {
      this.message = messageToSend;
      this.hasMessage = true;
      this.lock.notifyAll();
    }
  }

  public void setMessage(final Message messageToSend) {
    final Transaction tx = Environment.getCurrent().get(Transaction.class);
    if (tx != null) {
      final Synchronization synchro = new Synchronization() {

        private final Message msg = messageToSend.duplicate();

        public void beforeCompletion() {
          if (MessageCarrier.this.hasMessage) {
            throw new OrchestraRuntimeException("Message already associated");
          }
        }
        public void afterCompletion(final int status) {
          if (status == javax.transaction.Status.STATUS_COMMITTED) {
            MessageCarrier.this.replyMessage(this.msg);
          }
        }
      };
      tx.registerSynchronization(synchro);
    } else {
      this.replyMessage(messageToSend);
    }
  }

  public boolean hasMessage() {
    synchronized (this.lock) {
      return this.hasMessage;
    }
  }

  public String getUuid() {
    return this.uuid;
  }

}
