/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.var.Message;


/**
 * @author Guillaume Porcher
 *
 */
public class ReceivingService {
  /**
   * Handle incoming message for the given operation key. Use the given
   * {@link MessageCarrier} to store the reply.
   *
   * @param incomingMessage message received
   * @param operationKey operation key of the service that received the message
   * @param messageCarrier object that will be used to get the response
   * @return the execution to which the message was given. <code>null</code>
   *         if there is no execution waiting for this message (the message is
   *         stored).
   */
  public static BpelExecution handle(final Message incomingMessage,
      final OperationKey operationKey,
      final MessageCarrier messageCarrier) {
    final CommandService commandService = EnvFactoryRepository.get().get(CommandService.class);

    ExecElementToSignal execToSignal = null;
    // Step 1 : give the message to an existing instance or create a new
    // instance
    // This step is synchronized on the process
    execToSignal = commandService.execute(
        new AssociateMessageCommand(operationKey, incomingMessage, messageCarrier)
    );
    // Step 2 : continue execution
    // Synchronize on the instance to prevent concurrent modification in
    // separate transactions
    if (execToSignal != null) {
      while (!commandService.execute(new SignalExecCommand(execToSignal))) {
        // redistribute message
        execToSignal = commandService.execute(new ReAssociateMessageCommand(execToSignal));
        if (execToSignal == null) {
          return null;
        }
      }
      if (execToSignal != null) {
        commandService.execute(new RemovePendingMessageCommand(execToSignal.getPendingMessage()));
      }
      final BpelExecution currentInstance = execToSignal.getBpelExecution().getProcessInstance();

      ExecElementToSignal execElt = execToSignal;
      // Step 3 : for each child execution, verify that there is no message
      // available.
      // This step is synchronized on the process as we modify the receiver
      // (shared between all instances of a process)
      while (execElt != null && !execElt.getBpelExecution().isEnded()) {
        execElt = commandService.execute(new FindExecutionToSignalCommand(currentInstance));
        // Step 4: if a message can be associated with an execution, execute
        // the instance
        // (as for step 2, this is synchronized on the instance)
        while (execElt != null && !commandService.execute(new SignalExecCommand(execElt))) {
          // redistribute message
          execElt = commandService.execute(new ReAssociateMessageCommand(execElt));
        }
        if (execElt != null) {
          commandService.execute(new RemovePendingMessageCommand(execElt.getPendingMessage()));
        }
      }
      return execToSignal.getBpelExecution();
    }

    return null;
  }



}
