/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Map.Entry;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

import javax.wsdl.Part;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.StackObjectPool;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.internal.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;

/**
 * @author Charles Souillard
 */
public final class BpelUtil {

  private static DocumentBuilderFactory documentBuilderFactory;
  private static TransformerFactory transformerFactory;
  private static ObjectPool transformerPool = new StackObjectPool(new PoolTransformerFactory());;
  private static ObjectPool documentBuilderPool = new StackObjectPool(new PoolDocumentBuilderFactory());

  /**
   * Default private constructor to prevent this class to be instanciated.
   */
  private BpelUtil() { }

  private static DocumentBuilderFactory getDocumentBuilderFactory() {
    if (BpelUtil.documentBuilderFactory == null) {
      BpelUtil.documentBuilderFactory = DocumentBuilderFactory.newInstance();
    }
    return BpelUtil.documentBuilderFactory;
  }
  private static TransformerFactory getTransformerFactory() {
    if (BpelUtil.transformerFactory == null) {
      BpelUtil.transformerFactory = TransformerFactory.newInstance();
    }
    return BpelUtil.transformerFactory;
  }

  private static class PoolTransformerFactory extends BasePoolableObjectFactory {
    @Override
    public Object makeObject() {
      try {
        return BpelUtil.getTransformerFactory().newTransformer();
      } catch (final TransformerConfigurationException e) {
        throw new OrchestraRuntimeException(e);
      }
    }
    //nothing to do when passivating
    //public void passivateObject(Object obj);
  }
  private static class PoolDocumentBuilderFactory extends BasePoolableObjectFactory {
    @Override
    public Object makeObject() {
      try {
        final DocumentBuilderFactory myDocumentBuilderFactory = BpelUtil.getDocumentBuilderFactory();
        myDocumentBuilderFactory.setNamespaceAware(true);
        myDocumentBuilderFactory.setExpandEntityReferences(false);
        return myDocumentBuilderFactory.newDocumentBuilder();
      } catch (final ParserConfigurationException e) {
        throw new OrchestraRuntimeException(e);
      }
    }
    @Override
    public void passivateObject(final Object obj) {
      ((DocumentBuilder)obj).reset();
    }
  }
  private static Transformer getTransformer() {
    try {
      return (Transformer) BpelUtil.transformerPool.borrowObject();
    } catch (final Exception e) {
      throw new OrchestraRuntimeException(e);
    }
  }
  private static void returnTransformer(final Transformer transformer) {
    if (transformer != null) {
      try {
        BpelUtil.transformerPool.returnObject(transformer);
      } catch (final Exception e) {
        throw new OrchestraRuntimeException(e);
      }
    }
  }
  private static DocumentBuilder getDocumentBuilder() {
    try {
      return (DocumentBuilder) BpelUtil.documentBuilderPool.borrowObject();
    } catch (final Exception e) {
      throw new OrchestraRuntimeException(e);
    }
  }
  private static void returnDocumentBuilder(final DocumentBuilder documentBuilder) {
    if (documentBuilder != null) {
      try {
        BpelUtil.documentBuilderPool.returnObject(documentBuilder);
      } catch (final Exception e) {
        throw new OrchestraRuntimeException(e);
      }
    }
  }
  public static Document getNewDocument() {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      return documentBuilder.newDocument();
    } catch (final Exception e) {
      throw new OrchestraRuntimeException(e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  /**
   * This method writes a DOM document to a file.
   * @param doc - doc to write
   * @param file - dest file;
   */
  public static void writeXmlFile(final Document doc, final File file) {
    Transformer transformer = null;
    FileOutputStream fos = null;
    try {
      // Prepare the DOM document for writing
      final Source source = new DOMSource(doc);
      fos = new FileOutputStream(file);
      final Result result = new StreamResult(fos);

      // Write the DOM document to the file
      transformer = BpelUtil.getTransformer();
      transformer.transform(source, result);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception caught while writing xml file", e);
    } finally {
      if (fos != null) {
        try {
          fos.close();
        } catch (final Exception e) {
          // nothing
        }
      }
      BpelUtil.returnTransformer(transformer);
    }
  }

  /**
   * Return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   * @param element dom element to analyse
   * @return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   */
  public static boolean isCreateInstance(final Element element) {
    final String createInstance = XmlUtil.attribute(element, "createInstance");
    if (createInstance != null && createInstance.equals("yes")) {
      return true;
    }
    return false;
  }
  /**
   * Get all namespaces defined in the given node and all of its parents.
   * @param n node from where namespaces are taken
   * @return all namespaces defined in the given node and all of its parents.
   */
  public static Map<String, String> getAllNameSpaces(final org.w3c.dom.Node n) {
    final HashMap<String, String> result = new HashMap<String, String>();
    if (n == null) {
      return result;
    }
    //we look for the NameSpace in the parent Node
    if (n.getParentNode() != null) {
      final Map<String, String> resultParent = BpelUtil.getAllNameSpaces(n.getParentNode());
      for (final Entry<String, String> ns : resultParent.entrySet()) {
        if (!result.containsKey(ns.getKey())) {
          result.put(ns.getKey(), ns.getValue());
        }
      }

    }
    final NamedNodeMap attributes = n.getAttributes();
    if (attributes != null) {
      for (int i = 0; i < attributes.getLength(); i++) {
        if (attributes.item(i).getNodeName().contains("xmlns:")) {
          final String nodeName = attributes.item(i).getNodeName();
          final String prefix = nodeName.split(":")[1];
          if (!result.containsKey(prefix)) {
            result.put(prefix, attributes.item(i).getTextContent());
          }
        }
      }
    }
    result.put(XmlConstants.BPEL_FUNC_PREFIX, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE);
    return result;
  }

  /**
   * Return a string representation of the given Node based on javax.xml.transform api.
   * @param n node to convert to a string
   * @return a string representation of the given Node based on javax.xml.transform api.
   */
  public static String getNodeAsString(final Node n) {
    Transformer transformer = null;
    try {
      final Source source = new DOMSource(n);
      final StringWriter stringWriter = new StringWriter();
      final Result res = new StreamResult(stringWriter);
      transformer = BpelUtil.getTransformer();
      transformer.transform(source, res);
      //return stringWriter.getBuffer().toString();

      final String resultAsString = stringWriter.getBuffer().toString();
      final String prolog = "<?xml";
      if (resultAsString.startsWith(prolog)) {
        final String prologEnd = "?>";
        final int prologEndIndex = resultAsString.indexOf(prologEnd);
        return resultAsString.substring(prologEndIndex + prologEnd.length());
      }
      return resultAsString;
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception caugth when transforming node to string", e);
    } finally {
      BpelUtil.returnTransformer(transformer);
    }
  }

  public static ByteArrayOutputStream getNodeAsBAOS(final Node n) {
    Transformer transformer = null;
    try {
      final Source source = new DOMSource(n);
      final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      final Result res = new StreamResult(byteStream);
      transformer = BpelUtil.getTransformer();

      transformer.transform(source, res);

      return byteStream;
    } catch (final ClassCastException e) {
      throw new OrchestraRuntimeException(e);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception caugth when transforming node to bytestream", e);
    } finally {
      BpelUtil.returnTransformer(transformer);
    }
  }

  public static ByteArrayOutputStream getNodeAsBAOS(final Node n, final Integer deflateLevel) {
    Transformer transformer = null;
    OutputStream deflaterStream = null;
    try {
      final Source source = new DOMSource(n);
      final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      Result res = null;
      if (deflateLevel == null) {
        res = new StreamResult(byteStream);
      } else {
        final Deflater deflater = new Deflater(deflateLevel.intValue());
        deflaterStream = new DeflaterOutputStream(byteStream, deflater);
        res = new StreamResult(deflaterStream);
      }

      transformer = BpelUtil.getTransformer();
      transformer.transform(source, res);


      return byteStream;
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception caugth when transforming node to string", e);
    } finally {
      if (deflaterStream != null) {
        try {
          deflaterStream.close();
        } catch (final Exception e) {
          // nothing
        }
      }
      BpelUtil.returnTransformer(transformer);
    }
  }

  /**
   * Returns the most local NameSpace corresponding to given prefix.
   * @param n : the child from where to begin the search
   * @param prefix : prefix linking to searched namespaceURI
   * @return the most local NameSpace corresponding to given prefix
   */
  public static String getNamespaceURI(final org.w3c.dom.Node n, final String prefix) {
    final Node prefixDeclaration = n.getAttributes().getNamedItem("xmlns:" + prefix);
    if (prefixDeclaration != null) {
      //we have found the good NameSpace
      return prefixDeclaration.getNodeValue();
    }
    // we have found the good NameSpace
    //we look for the NameSpace in the parent Node
    return BpelUtil.getNamespaceURI(n.getParentNode(), prefix);
  }

  /**
   * remove the prefix of the String.
   * @param s the String which has to be cut
   * @return the string without the prefix
   */
  public static String removePrefix(final String s) {
    int i = s.indexOf(":");
    return s.substring(++i);
  }

  /**
   * Look for an instance of a class in a list of ExtensibilityElement.
   * @param listOfExtensibilityElement the list of ExtensibilityElement.
   * @param classToFind : the class we llok for
   * @return If found, return the element found, else, return null.
   */
  @SuppressWarnings("unchecked")
  public static Object findExtensibilityElement(final List listOfExtensibilityElement, final Class classToFind) {
    for (final Object extElem : listOfExtensibilityElement) {
      if (classToFind.isInstance(extElem)) {
        return extElem;
      }
    }
    return null;
  }
  /**
   * return true if the given ns is a xmlSchema.
   * @param ns ns to analyse
   * @return true if the given ns is a xmlSchema.
   */
  public static boolean isXmlSchemaNs(final String ns) {
    return ns != null && (ns.equals(XmlConstants.XMLNS_XMLSCHEMA) || ns.equals(XmlConstants.XMLNS_XMLSCHEMAXSD));
  }
  /**
   * Return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   * @param address address to use to build the serviceRef document
   * @param serviceQN serviceQN to use to build the serviceRef document
   * @return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   */
  public static Document getServiceRefDocument(final SOAPAddress address, final QName serviceQN) {
    final DocumentBuilder documentBuilder = BpelUtil.getDocumentBuilder();
    final Document document = documentBuilder.newDocument();
    BpelUtil.returnDocumentBuilder(documentBuilder);

    final Element serviceRef = document.createElementNS(XmlConstants.XMLNS_BPEL_2_0_SERVICEREF, XmlConstants.SERVICE_REF_TAG_NAME);
    document.appendChild(serviceRef);

    final Element epr = document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    serviceRef.appendChild(epr);

    final Element addressElement = document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE,XmlConstants.ADDRESS_TAG_NAME);
    addressElement.setTextContent(address.getLocationURI());
    epr.appendChild(addressElement);

    final Element serviceNameElement = document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE,XmlConstants.SERVICE_NAME_TAG_NAME);
    serviceNameElement.setAttribute("xmlns:eprServicePrefix", serviceQN.getNamespaceURI());
    serviceNameElement.setTextContent("eprServicePrefix:" + serviceQN.getLocalPart());
    epr.appendChild(serviceNameElement);

    return document;
  }

  public static List<Element> getChildrenElements(final Node node) {
    final List<Element> elements = new ArrayList<Element>();
    final NodeList children = node.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      final Node child = children.item(i);
      if (child instanceof Element) {
        elements.add((Element)child);
      }
    }
    return elements;
  }

  public static Document getDocumentFromString(String s) {
    final String prolog = "<?xml";
    if (s.startsWith(prolog)) {
      final String prologEnd = "?>";
      final int prologEndIndex = s.indexOf(prologEnd);
      s = s.substring(prologEndIndex + prologEnd.length());
    }
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      return documentBuilder.parse(new InputSource(new StringReader(s)));
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Esception while building a document from string : " + s, e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  public static Document getDocumentFromInputStream(final InputStream is) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      return documentBuilder.parse(is);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception while building a document from input stream : " + is, e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  public static Document getDocumentFromReader(final Reader r) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      return documentBuilder.parse(new InputSource(r));
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception while building a document from reader : " + r, e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  public static List<Element> getElements(final Node n) {
    final NodeList nodeList = n.getChildNodes();
    final List<Element> elements = new ArrayList<Element>();
    for (int i = 0;i < nodeList.getLength();i++) {
      if (nodeList.item(i) instanceof Element) {
        elements.add((Element)nodeList.item(i));
      }
    }
    return elements;
  }

  public static List<Text> getTexts(final Node n) {
    final NodeList nodeList = n.getChildNodes();
    final List<Text> texts = new ArrayList<Text>();
    for (int i = 0;i < nodeList.getLength();i++) {
      if (nodeList.item(i) instanceof Text) {
        texts.add((Text)nodeList.item(i));
      }
    }
    return texts;
  }
  public static Document getDocumentFromFile(final File xmlFile) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      return documentBuilder.parse(xmlFile);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Unable to parse successfully xml file : " + xmlFile, e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  public static Document getDocumentFromURL(final URL xmlUrl) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = BpelUtil.getDocumentBuilder();
      final InputStream inStream = xmlUrl.openStream();
      return documentBuilder.parse(inStream);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Unable to parse successfully xml url : " + xmlUrl, e);
    } finally {
      BpelUtil.returnDocumentBuilder(documentBuilder);
    }
  }

  public static Element getDocumentWithOneElement(final QName elementQName) {
    final Document document = BpelUtil.getNewDocument();
    final Element rootElement = document.createElementNS(elementQName.getNamespaceURI(), elementQName.getLocalPart());
    document.appendChild(rootElement);
    return rootElement;
  }

  public static Element getPartElement(final Element father, final Part part) {
    final String partName = part.getName();
    final List<Element> childrenElements = BpelUtil.getElements(father);
    for (final Element child : childrenElements) {
      if (child.getLocalName().equals(partName)) {
        return child;
      }
    }
    return null;
  }

  public static int getRandomInt(final long diff, final int max) {
    final Random generator = new Random(System.currentTimeMillis() + diff);
    return generator.nextInt(max);
  }

  public static Boolean booleanValue(final String text, final boolean defaultValue) {
    final Boolean result = BpelUtil.booleanValue(text);
    if (result == null) {
      return defaultValue;
    }
    return result;
  }

  public static Boolean booleanValue(final String text) {
    if (text == null) {
      return null;
    }

    // if we have to check for value true
    if (("true".equalsIgnoreCase(text))
        || ("enabled".equalsIgnoreCase(text))
        || ("on".equalsIgnoreCase(text))
        || ("yes".equalsIgnoreCase(text))
    ) {
      return Boolean.TRUE;
    }
    if (("false".equalsIgnoreCase(text))
        || ("disabled".equalsIgnoreCase(text))
        || ("off".equalsIgnoreCase(text))
        || ("no".equalsIgnoreCase(text))
    ) {
      return Boolean.FALSE;
    }

    return null;
  }

  /**
   * Test the equality of two nodes following same conditions that org.w3c.dom.Node.isEqualNode(Node n) except for prefix.
   * The two nodes prefix doesn't need to be equals.
   * @param n1 First node
   * @param n2 Second node
   * @return true if the two nodes are equals, false otherwise
   */
  public static Boolean areNodeEquals(final Node n1, final Node n2) {
    Misc.checkArgsNotNull(n1, n2);

    Node n1Temp, n2Temp;
    int i = 0;
    NamedNodeMap n1NamedNodeMap;
    NamedNodeMap n2NamedNodeMap;

    if (n1.getNodeType() != n2.getNodeType()) {
      return false;
    }
    if (n1.getNodeType() == Node.DOCUMENT_TYPE_NODE && !BpelUtil.areDocumentTypeEquals((DocumentType) n1, (DocumentType) n2)) {
      return false;
    }

    if (!BpelUtil.areStringEquals(n1.getLocalName(), n2.getLocalName())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(n1.getNamespaceURI(), n2.getNamespaceURI())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(n1.getNodeValue(), n2.getNodeValue())) {
      return false;
    }
    n1NamedNodeMap = n1.getAttributes();
    n2NamedNodeMap = n2.getAttributes();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }

    String n1Name, n2Name;
    boolean found = false;
    int j = 0;
    //Iterate the first NamedNodeMap and verify that the elements exist in the second NamedNodeMap
    if (n1NamedNodeMap != null && n1NamedNodeMap.getLength() > 0) {
      for (i = 0;i < n1NamedNodeMap.getLength();i++) {
        n1Temp = n1NamedNodeMap.item(i);
        if (n1Temp.getNodeName().contains(":")) {
          n1Name = null;
        } else {
          n1Name = n1Temp.getNodeName();
        }
        n1Temp.normalize();
        for (j = 0;!found && j < n2NamedNodeMap.getLength();j++) {
          n2Temp = n2NamedNodeMap.item(j);
          if (n2Temp.getNodeName().contains(":")) {
            n2Name = null;
          } else {
            n2Name = n2Temp.getNodeName();
          }
          n2Temp.normalize();
          found = n1Temp.getNodeValue().equals(n2Temp.getNodeValue()) && BpelUtil.areStringEquals(n1Name, n2Name)
            && BpelUtil.areStringEquals(n1Temp.getNamespaceURI(), n2Temp.getNamespaceURI());
        }
        if (!found) {
          return false;
        }
      }
    }

    final NodeList n1ChildNodes = n1.getChildNodes();
    final NodeList n2ChildNodes = n2.getChildNodes();
    if ((n1ChildNodes != null && n2ChildNodes == null) || (n1ChildNodes == null && n2ChildNodes != null)) {
      return false;
    }
    if (n1ChildNodes != null && n2ChildNodes != null) {
      if (n1ChildNodes.getLength() != n2ChildNodes.getLength()) {
        return false;
      }
      for (i = 0;i < n1ChildNodes.getLength();i++) {
        n1Temp = n1ChildNodes.item(i);
        n2Temp = n2ChildNodes.item(i);
        if (n2Temp == null) {
          return false;
        }
        n1Temp.normalize();
        n2Temp.normalize();
        if (!BpelUtil.areNodeEquals(n1Temp, n2Temp)) {
          return false;
        }
      }
    }
    return true;
  }

  //Supplementary checks in case of DocumentType Nodes
  private static boolean areDocumentTypeEquals(final DocumentType documentType1, final DocumentType documentType2) {
    NamedNodeMap n1NamedNodeMap;
    NamedNodeMap n2NamedNodeMap;
    if (!BpelUtil.areStringEquals(documentType1.getPublicId(), documentType2.getPublicId())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(documentType1.getSystemId(), documentType2.getSystemId())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(documentType1.getInternalSubset(), documentType2.getInternalSubset())) {
      return false;
    }
    n1NamedNodeMap = documentType1.getEntities();
    n2NamedNodeMap = documentType2.getEntities();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    if (!BpelUtil.areNamedNodeMapEquals(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }

    n1NamedNodeMap = documentType1.getNotations();
    n2NamedNodeMap = documentType2.getNotations();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    if (!BpelUtil.areNamedNodeMapEquals(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    return true;
  }

  //Return true if the two strings are null or equals, false otherwise
  private static boolean areStringEquals(final String str1, final String str2) {
    if (str1 == null && str2 == null) {
      return true;
    }
    if (str1 == null) {
      return false;
    }
    if (str2 == null) {
      return false;
    }
    return str1.equals(str2);
  }

  //Return true if the two NamedNodeMap are null or have same length, false otherwise
  private static boolean areNamedNodeMapNullOrDiffLength(final NamedNodeMap n1NamedNodeMap, final NamedNodeMap n2NamedNodeMap) {
    if ((n1NamedNodeMap != null && n1NamedNodeMap == null) || (n1NamedNodeMap == null && n2NamedNodeMap != null)) {
      return false;
    }
    if (n1NamedNodeMap != null && n2NamedNodeMap != null && n1NamedNodeMap.getLength() != n2NamedNodeMap.getLength()) {
      return false;
    }
    return true;
  }

  //Return true if the two NamedNodeMap are equals : equals element (using isEqualNode) at the same index, false otherwise
  private static boolean areNamedNodeMapEquals(final NamedNodeMap n1NamedNodeMap, final NamedNodeMap n2NamedNodeMap) {
    if (n1NamedNodeMap != null && n1NamedNodeMap.getLength() > 0) {
      Node n1Temp, n2Temp;
      int i = 0;
      for (i = 0;i < n1NamedNodeMap.getLength();i++) {
        n1Temp = n1NamedNodeMap.item(i);
        n2Temp = n2NamedNodeMap.item(i);
        if (n2Temp == null) {
          return false;
        }
        n1Temp.normalize();
        n2Temp.normalize();
        if (!n1Temp.isEqualNode(n2Temp)) {
          return false;
        }
      }
    }
    return true;
  }
}


