/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.MimeHeaders;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;
import javax.xml.transform.dom.DOMSource;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

public final class SOAPUtil {

  private static SOAPConnectionFactory conFactory;
  private static MessageFactory messagefactory;

  private SOAPUtil() {

  }

  private static synchronized SOAPConnectionFactory getSOAPConnectionFactory() {
    if (SOAPUtil.conFactory == null) {
      try {
        SOAPUtil.conFactory = SOAPConnectionFactory.newInstance();
      } catch (final Exception e) {
        throw new OrchestraRuntimeException(e);
      }
    }
    return SOAPUtil.conFactory;
  }

  private static synchronized MessageFactory getMessageFactory() {
    if (SOAPUtil.messagefactory == null) {
      try {
        SOAPUtil.messagefactory = MessageFactory.newInstance();
      } catch (final Exception e) {
        throw new OrchestraRuntimeException(e);
      }
    }
    return SOAPUtil.messagefactory;
  }

  private static SOAPMessage getSOAPMessageFromDocument(final Document document) {
    SOAPMessage soapMessage = null;
    try {
      soapMessage = SOAPUtil.getMessageFactory().createMessage();
      final SOAPPart soapPart = soapMessage.getSOAPPart();
      soapPart.setContent(new DOMSource(document));
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception when creating a SOAPMessage from document : " + document, e);
    }
    return soapMessage;
  }

  public static SOAPMessage buildSOAPMessage(final Document document, final String soapAction) {
    SOAPMessage soapMessage = null;
    try {
      soapMessage = SOAPUtil.getMessageFactory().createMessage();

      final MimeHeaders hd = soapMessage.getMimeHeaders();
      hd.addHeader("SOAPAction", soapAction);
      final SOAPPart soapPart = soapMessage.getSOAPPart();

      final SOAPEnvelope enveloppe = soapPart.getEnvelope();
      enveloppe.addNamespaceDeclaration("env", "http://schemas.xmlsoap.org/soap/envelope/");

      final SOAPBody body = enveloppe.getBody();

      body.addDocument(document);

      soapMessage.saveChanges();
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception while building a SOAPMessage", e);
    }
    return soapMessage;
  }

  public static SOAPMessage buildRpcSOAPMessage(final String soapAction, final Message message, final String operationName) {
    final Document document = SOAPUtil.buildRpcDocument(message, operationName);
    return SOAPUtil.buildSOAPMessage(document, soapAction);
  }

  public static Document buildRpcDocument(final Message message, final String operationName) {
    Document document = null;
    try {
      document = BpelUtil.getNewDocument();
      final Element operation = document.createElement(operationName);

      final Iterator<Entry<String, Element>> it = message.getParts().entrySet().iterator();
      while (it.hasNext()) {
        final Entry<String, Element> entry = it.next();
        final String partName = entry.getKey();
        final Element partElement = entry.getValue();

        if (BpelXmlUtil.isDefaultElement(partElement)) {
          final Element partEl = document.createElement(partName);
          partEl.setTextContent(partElement.getTextContent());
          operation.appendChild(partEl);
        } else {
          final Node importedPart = document.importNode(partElement, true);
          operation.appendChild(importedPart);
        }
      }
      document.appendChild(operation);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception while building Rpc SOAP Message", e);
    }
    return document;
  }

  public static SOAPMessage buildDocumentSOAPMessage(final String soapAction, final Message message) {
    final Element part = message.getParts().values().iterator().next();
    return SOAPUtil.buildDocumentSOAPMessage(soapAction, part);
  }

  public static SOAPMessage buildDocumentSOAPMessage(final String soapAction, final Element part) {
    Document document = null;
    try {
      document = BpelUtil.getNewDocument();
      final Node importedPart = document.importNode(part, true);
      document.appendChild(importedPart);

    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception while building Rpc SOAP Message", e);
    }
    return SOAPUtil.buildSOAPMessage(document, soapAction);
  }

  public static SOAPMessage buildDocumentSOAPMessage(final String soapAction, final String part) {
    final Element partEl = BpelUtil.getDocumentFromString(part).getDocumentElement();
    return SOAPUtil.buildDocumentSOAPMessage(soapAction, partEl);
  }

  public static SOAPMessage call(final SOAPMessage soapMessage, final String endpoint) {
    SOAPMessage response = null;
    SOAPConnection connection = null;
    try {
      connection = SOAPUtil.getSOAPConnectionFactory().createConnection();
      response = connection.call(soapMessage, endpoint);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("Exception caught while performing call", e);
    } finally {
      if (connection != null) {
        try {
          connection.close();
        } catch (final Exception e) {
          // nothing
        }
      }
    }
    return response;
  }

  public static Message getMessageFromRpcElement(final Element operationElement, final Collection<Part> parts) {
    final HashMap <String, Document> messageParts = new HashMap<String, Document>();
    final Message message = new Message(messageParts);

    for (final Part part : parts) {
      final Element el = BpelUtil.getPartElement(operationElement, part);
      if (part.getTypeName() != null && BpelUtil.isXmlSchemaNs(part.getTypeName().getNamespaceURI())) {
        //it is a schema type
        // TODO : manage simpleTypes here
        final String text = ((Text)el.getFirstChild()).getData();
        final Element element = BpelXmlUtil.createElementWithContent(text);
        messageParts.put(part.getName(), element.getOwnerDocument());
      } else {
        final Document document = BpelUtil.getNewDocument();
        final Node node = document.importNode(el,true);
        document.appendChild(node);
        messageParts.put(part.getName(), document);
      }
    }
    return message;
  }

  public static Message buildMessageFromRpcSOAPBodyElement(final Collection<Part> parts, final Element soapBodyElement) {
    final HashMap<String, Document> messageParts = new HashMap<String, Document>();
    for (final Part part : parts) {
      final Element el = BpelUtil.getPartElement(soapBodyElement, part);
      if (el == null) {
        throw new OrchestraRuntimeException("Impossible to build a Message from the given rpc soap body element."
            + " One element with a localName equals to " + part.getName() + " is mandatory.");
      }
      if (part.getTypeName() != null && BpelUtil.isXmlSchemaNs(part.getTypeName().getNamespaceURI())) {
        //it is a schema type
        String text = null;
        final Node child = el.getFirstChild();
        if (child != null) {
          text = ((Text)child).getData();
        } else {
          text = el.getTextContent();
        }

        final Element element = BpelXmlUtil.createElementWithContent(text);
        messageParts.put(part.getName(), element.getOwnerDocument());
      } else {
        final Document document = BpelUtil.getNewDocument();
        final Node node = document.importNode(el,true);
        document.appendChild(node);
        messageParts.put(part.getName() , document);
      }
    }
    final Message message = new Message(messageParts);
    return message;
  }
  public static Message buildMessageFromDocumentSOAPBodyElement(final Collection<Part> parts, final Element soapBodyElement) {
    final HashMap <String, Document> messageParts = new HashMap<String, Document>();
    final Part part = parts.iterator().next();
    final String partName = part.getName();
    final Document document = BpelUtil.getNewDocument();
    final Node node = document.importNode(soapBodyElement, true);
    document.appendChild(node);
    messageParts.put(partName, document);
    return new Message(messageParts);
  }

}

