/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.wsdl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.wsdl.Definition;
import javax.wsdl.Message;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.factory.WSDLFactory;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;

/**
 *
 * @author Charles Souillard
 */
public class WsdlsInfos implements Serializable {

  /**   */
  private static final long serialVersionUID = 2688615599128088745L;

  protected long dbid;

  protected Definition wsdlDefinition;

  /** partnerLinkTypes of the corresponding wsdl. */
  protected Map<QName, PartnerLinkType> partnerLinkTypes;

  /** properties of the corresponding wsdl. */
  protected Map<QName, Property> properties;

  /** propertyAliases of the corresponding wsdl. */
  protected List<PropertyAlias> propertyAliases;

  /**
   * Default constructor.
   */
  public WsdlsInfos() {
    try {
      this.wsdlDefinition = WSDLFactory.newInstance().newDefinition();
    } catch (final WSDLException e) {
      throw new OrchestraRuntimeException(e);
    }
  }

  /**
   * Add a PartnerLinkType.
   *
   * @param partnerLinkType
   *            partnerLinkType to add
   */
  public void addPartnerLinkType(final PartnerLinkType partnerLinkType) {
    if (this.partnerLinkTypes == null) {
      this.partnerLinkTypes = new HashMap<QName, PartnerLinkType>();
    }
    this.partnerLinkTypes.put(partnerLinkType.getName(), partnerLinkType);
  }

  /**
   * Return the partnerLinkType having the given QName.
   *
   * @param partnerLinkTypeName
   *            partnerLinkTypeName
   * @return the partnerLinkType having the given QName.
   */
  public PartnerLinkType getPartnerLinkType(final QName partnerLinkTypeName) {
    if (this.partnerLinkTypes != null) {
      return this.partnerLinkTypes.get(partnerLinkTypeName);
    }
    return null;
  }

  /**
   * Add all portTypes.
   *
   * @param portTypesToAdd
   *            portTypes to add
   */
  public void addPortTypes(final Map<QName, PortType> portTypesToAdd) {
    for (final PortType pt : portTypesToAdd.values()) {
      this.wsdlDefinition.addPortType(pt);
    }
  }

  /**
   * Return the portType having the given QName.
   *
   * @param portTypeQName
   *            portTypeName
   * @return the portType having the given QName.
   */
  public PortType getPortType(final QName portTypeQName) {
    return this.wsdlDefinition.getPortType(portTypeQName);
  }

  /**
   * Add all messages.
   *
   * @param messagesToAdd
   *            messages to add
   */
  public void addMessages(final Map<QName, Message> messagesToAdd) {
    for (final Message msg : messagesToAdd.values()) {
      this.wsdlDefinition.addMessage(msg);
    }
  }

  /**
   * Return the message having the given QName.
   *
   * @param messageName
   *            messageName
   * @return the message having the given QName.
   */
  public Message getMessage(final QName messageName) {
    return this.wsdlDefinition.getMessage(messageName);
  }

  /**
   * Add all services.
   *
   * @param servicesToAdd
   *            services to add
   */
  public void addServices(final Map<QName, Service> servicesToAdd) {
    for (final Service service : servicesToAdd.values()) {
      this.wsdlDefinition.addService(service);
    }
  }

  /**
   * Return the service having the given QName.
   *
   * @param serviceName
   *            serviceName
   * @return the service having the given QName.
   */
  public Service getService(final QName serviceName) {
    return this.wsdlDefinition.getService(serviceName);
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public Service getServiceOfPort(final Port port) {
    final String portName = port.getName();
    for (final Service service : (Collection<Service>) this.wsdlDefinition.getServices().values()) {
      if (service.getPort(portName) != null) {
        return service;
      }
    }
    return null;
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public List<Service> getServicesOfPortType(final QName portTypeQName) {
    final List<Service> matchinServices = new ArrayList<Service>();
    for (final Service service : (Collection<Service>) this.wsdlDefinition.getServices().values()) {
      for (final Port port : (Collection<Port>) service.getPorts().values()) {
        if (port.getBinding().getPortType().getQName().equals(portTypeQName)) {
          matchinServices.add(service);
        }
      }
    }
    if (matchinServices.size() > 0) {
      return matchinServices;
    }
    return null;
  }

  /**
   * Return all ports implementing the given PortType.
   *
   * @param portTypeQName
   *            portType
   * @return all ports implementing the given PortType
   */
  @SuppressWarnings("unchecked")
  public List<Port> getPorts(final QName portTypeQName) {
    final List<Port> ports = new ArrayList<Port>();
    for (final Service service : (Collection<Service>) this.wsdlDefinition.getServices().values()) {
      final Map<QName, Port> servicePorts = service.getPorts();
      for (final Port port : servicePorts.values()) {
        if (port.getBinding().getPortType().getQName().equals(portTypeQName)) {
          ports.add(port);
        }
      }
    }
    return ports;
  }

  /**
   * Add the given property.
   *
   * @param property
   *            property to add
   */
  public void addProperty(final Property property) {
    if (this.properties == null) {
      this.properties = new HashMap<QName, Property>();
    }
    this.properties.put(property.getName(), property);
  }

  /**
   * Return the property having the given QName.
   *
   * @param propertyQName
   *            propertyQName
   * @return the property having the given QName.
   */
  public Property getProperty(final QName propertyQName) {
    if (this.properties == null) {
      return null;
    }
    return this.properties.get(propertyQName);
  }

  /**
   * Add the given PrpertyAlias.
   *
   * @param propertyAlias
   *            propertyAlias to add
   */
  public void addPropertyAlias(final PropertyAlias propertyAlias) {
    if (this.propertyAliases == null) {
      this.propertyAliases = new ArrayList<PropertyAlias>();
    }
    this.propertyAliases.add(propertyAlias);
  }

  /**
   * Get the propertyAlias associated with the propertyName and the messageType
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param messageType :
   *            messageType of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithMessageType(final QName propertyName,
      final QName messageType) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Return all propertyAlias using the given messageType.
   *
   * @param messageType
   *            messageType
   * @return all propertyAlias using the given messageType.
   */
  public List<PropertyAlias> getPropertyAliasesWithMessageType(final QName messageType) {
    final List<PropertyAlias> matchingPropertyAliases = new ArrayList<PropertyAlias>();
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)) {
        matchingPropertyAliases.add(propertyAlias);
      }
    }
    return matchingPropertyAliases;
  }

  /**
   * Return all propertyAliases using the given messageType and properties.
   *
   * @param propertiesToScan
   *            properties we need to scan.
   * @param messageType
   *            messageType that returned propertyAlias should conform
   * @return all propertyAliases using the given messageType and properties.
   */
  public List<PropertyAlias> getPropertyAliases(
      final Collection<Property> propertiesToScan, final QName messageType) {
    final List<PropertyAlias> result = new ArrayList<PropertyAlias>();
    for (final Property property : propertiesToScan) {
      final PropertyAlias propertyAlias = this.getPropertyAliasWithMessageType(property
          .getName(), messageType);
      if (propertyAlias == null) {
        throw new OrchestraRuntimeException("propertyAlias is null, property = "
            + property);
      }
      result.add(propertyAlias);
    }
    return result;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the element
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param element :
   *            element of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithElement(final QName propertyName,
      final QName element) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getElement() != null
          && propertyAlias.getElement().equals(element)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the xmlType given
   * in parameter. Or null if there is no propertyAlias associated with this
   * propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param xmlType :
   *            xmlType of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithType(final QName propertyName, final QName type) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getType() != null
          && propertyAlias.getType().equals(type)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }
}
